/*
 * Copyright (C) 2008 Intel Corporation. All Rights Reserved.
 *
 * The source code contained or described herein and all
 * documents related to the source code ("Material") are owned by
 * Intel Corporation or its suppliers or licensors. Title to the
 * Material remains with Intel Corporation or its suppliers and
 * licensors. The Material is protected by worldwide copyright
 * laws and treaty provisions.  No part of the Material may be
 * used, copied, reproduced, modified, published, uploaded,
 * posted, transmitted, distributed,  or disclosed in any way
 * except as expressly provided in the license provided with the
 * Materials.  No license under any patent, copyright, trade
 * secret or other intellectual property right is granted to or
 * conferred upon you by disclosure or delivery of the Materials,
 * either expressly, by implication, inducement, estoppel or
 * otherwise, except as expressly provided in the license
 * provided with the Materials
*/
// [DESCRIPTION]
// 
// Solve the nqueens problem STL version
//
// [COMPILE]  
// Compile on IA-32 and Intel(R) 64 architectures only. 
// Use the following compiler options to compile: 
//
// CommandLine: icl /DWIN32 nq-stl.cpp winmm.lib
//  


#include <fstream>
#include <iostream>
#include <vector>

using namespace std;

int nrOfSolutions=0;
int size=0;

// ------------------- Timing ----------------------

//****************************************************
//*  Windows NT (32 bit) dtime() routine             *
//*  Provided by: Piers Haken, piersh@microsoft.com  *
//****************************************************
#ifdef WIN32
#include <windows.h>
#include <mmsystem.h>

double dtime(void)
{
 double q;

 q = (double)GetTickCount() * 1.0e-03;

 return q;
}
#endif

//*****************************************************
//*  UNIX dtime(). This is the preferred UNIX timer.  *
//*****************************************************
#ifdef UNIX
#include <sys/time.h>
#include <sys/resource.h>

#ifdef hpux
#include <sys/syscall.h>
#define getrusage(a,b) syscall(SYS_getrusage,a,b)
#endif

struct rusage rusage;

double dtime()
{
 double q;

 getrusage(RUSAGE_SELF,&rusage);

 q = (double)(rusage.ru_utime.tv_sec);
 q = q + (double)(rusage.ru_utime.tv_usec) * 1.0e-06;

 return q;
}
#endif

// ------------------- Timing ----------------------

void print(vector<int> &queens, ofstream &fout){
  fout << "Queens positions on board:" << endl ;
  for(short i=0; i<queens.size();i++){
    fout << queens[i]+1 << " ";
  }
  fout << endl;
}
int main(int argc, char*argv[]) {
  
  int column, i;
  double startTime, endTime, dtime();
  vector<int> queens;
  bool position;
  
  if(argc !=2) {
    cerr << "Usage: nq-stl boardSize.\n";
    return 0;
  }
  if (atoi(argv[1]) == 1){
    cerr << "Solution is trivial, no need for a computer :) " << endl;
    return 0;
  }
  else if (atoi(argv[1]) == 2 || atoi(argv[1]) == 3){
    cerr << "No solution, nice try ;) " << endl;
    return 0;
  }
  else{ 
    size = atoi(argv[1]);
  }

    cout << "Starting serial STL recursive solver for size " << size << "...\n";
#ifdef _DEBUG
  ofstream fout("nqueens.out");
  fout << "Starting algorithm trace" << endl ;
#endif  
  
  startTime = dtime();
  DWORD startTime1=timeGetTime();

  // Start algorithm

  queens.push_back(-1);
    
  while(!queens.empty()){
    column = ((int)queens.size())-1;
#ifdef _DEBUG
    fout << "column: " << column << endl ;
#endif
    position = false;
    // Loop as long as positions are ok and number of found columns is less than n
    while(!position && (queens[column])<(size-1)){
      queens[column] = queens[column] + 1 ; 
      position = true;
#ifdef _DEBUG
      fout << "queens[" << column << "]: " << queens[column] << endl ;
#endif
      for(i=0; i<column; i++){
        // Test if position cannot be attacked by queen from another row or diagonale
        if(queens[i]==queens[column] || abs(queens[i]-queens[column]) == column-i){
          position = false ;
        }
      }
#ifdef _DEBUG
      fout << "Location ok (1) ? = " << position << endl ;  
#endif
    }
    if(position){
      if(column == size-1){
        // Found one solution
#ifdef _DEBUG
        fout << endl << "!! Solution found  !!" << endl << endl ;  
        print(queens,fout);
#endif
        nrOfSolutions = nrOfSolutions + 1;
      }else{
        // Position valid. Try from here next column
#ifdef _DEBUG        
        fout << "++ Add value to vector in queens[]" << endl ;  
#endif
        queens.push_back(-1);
      }
    }else {
      // Position not valid go back and try again from previous column
#ifdef _DEBUG
      fout << "-- Remove one element in queens[]" << endl ;  
#endif
      queens.pop_back();
    }
  }

  // End algorithm
  endTime = dtime() - startTime;
  DWORD endTime1=timeGetTime();
  endTime = endTime;

#ifdef _DEBUG  
  fout << endl << endl ;   
  fout << "Solution for the Nqueens problem " << endl; 
  fout << "Board size: " << size << " :-: Solution(s): " << nrOfSolutions << endl;
  fout << "Calculations took " << endTime*1000.0 << " ms." << endl;
//  delete(fout);
  cout << "See file solution.out for algorithm trace! "<< endl ;
#endif  

  cout << "Number of solutions: " << nrOfSolutions << endl;
  cout << "Calculations took " << endTime1-startTime1 << " ms." << endl ;

  return 0;
}

  
