/*
 * Copyright (C) 2008 Intel Corporation. All Rights Reserved.
 *
 * The source code contained or described herein and all
 * documents related to the source code ("Material") are owned by
 * Intel Corporation or its suppliers or licensors. Title to the
 * Material remains with Intel Corporation or its suppliers and
 * licensors. The Material is protected by worldwide copyright
 * laws and treaty provisions.  No part of the Material may be
 * used, copied, reproduced, modified, published, uploaded,
 * posted, transmitted, distributed,  or disclosed in any way
 * except as expressly provided in the license provided with the
 * Materials.  No license under any patent, copyright, trade
 * secret or other intellectual property right is granted to or
 * conferred upon you by disclosure or delivery of the Materials,
 * either expressly, by implication, inducement, estoppel or
 * otherwise, except as expressly provided in the license
 * provided with the Materials
*/
// [DESCRIPTION]
// 
// Solve the nqueens problem tbb version
//
// [COMPILE]  
// Compile on IA-32 and Intel(R) 64 architectures only. 
// Use the following compiler options to compile: 
//
// CommandLine: icl /MD nq-tbb.cpp tbb.lib winmm.lib
//
// If building in Visual Studio, make sure that the Intel Threading Building
// Blocks include and lib folders are included in the paths used by the compiler.
// This is done by using the Project > Intel Build Components Selector tool.
// 
// Dependencies: Intel Threading Building Blocks (TBB) 
//

#include <iostream>
#include <windows.h>
#include <mmsystem.h>
#include "tbb/task_scheduler_init.h"
#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"
#include "tbb/spin_mutex.h"

using namespace tbb;
using namespace std;

int nrOfSolutions=0;
int size=0;

// define tbb mutexes
typedef spin_mutex NrOfSolutionsMutexType;
NrOfSolutionsMutexType NrOfSolutionsMutex;
typedef spin_mutex PrintMutexType;
PrintMutexType PrintMutex;


void print(int queens[]) {
  cerr << "Solution " << nrOfSolutions << endl; 
  for(int row=0; row<size; row++) {
    for(int col=0; col<size; col++) {
      if(queens[row]==col) {
        cout << "Q";
      }
      else {
        cout << "_";
      }
    }
    cout << endl;
  }
}

void setQueen(int queens[], int row, int col) {
  for(int i=0; i<row; i++) {
    // vertical attacks
    if (queens[i]==col) {
      return;
    }
    // diagonal attacks
    if (abs(queens[i]-col) == (row-i) ) {
      return;
    }
  }

  // column is ok, set the queen
  queens[row]=col;

  if(row==size-1) {
    {      
      // increment is not atomic, so setting a lock is required here
      NrOfSolutionsMutexType::scoped_lock mylock(NrOfSolutionsMutex);
      nrOfSolutions++;
    } // closing block invokes scoped_lock destructor which releases the lock
#ifdef _DEBUG
    {
      // only one thread is allowed to print at a time
      //PrintMutexType::scoped_lock mylock(PrintMutex);
      //print(queens);
    }
#endif
  }
  else {
    // try to fill next row (recursion)
    for(int i=0; i<size; i++) {
      setQueen(queens, row+1, i);
    }
  }
}

class SetQueens {
public:
  void operator()( const blocked_range<size_t>& r ) const {
    for( size_t i=r.begin(); i!=r.end(); ++i ) {
      // try all positions in first row
      // create separate array for each recursion
      // started here
      setQueen(new int[size], 0, (int)i);
    }
  }

};

void solve() {
  // Do a parallel for over the n positions in the first row.
  // Let the scheduler decide how the n tasks are distrubuted 
  // on the different threads
  parallel_for(blocked_range<size_t>(0, size, 1), SetQueens() );
}

int main(int argc, char*argv[]) {
  if(argc !=2) {
    cerr << "Usage: nq-tbb boardSize.\n";
    return 0;
  }

  size = atoi(argv[1]);
  cout << "Starting TBB solver for size " << size << "...\n";

  DWORD startTime=timeGetTime();

  // initialize tbb scheduler
  task_scheduler_init init;
  solve();
  DWORD endTime=timeGetTime();
  cout << "Number of solutions: " << nrOfSolutions << endl;
  cout << "Calculations took " << endTime-startTime << "ms.\n";

  return 0;
}