/*
 * Copyright (C) 2008 Intel Corporation. All Rights Reserved.
 *
 * The source code contained or described herein and all
 * documents related to the source code ("Material") are owned by
 * Intel Corporation or its suppliers or licensors. Title to the
 * Material remains with Intel Corporation or its suppliers and
 * licensors. The Material is protected by worldwide copyright
 * laws and treaty provisions.  No part of the Material may be
 * used, copied, reproduced, modified, published, uploaded,
 * posted, transmitted, distributed,  or disclosed in any way
 * except as expressly provided in the license provided with the
 * Materials.  No license under any patent, copyright, trade
 * secret or other intellectual property right is granted to or
 * conferred upon you by disclosure or delivery of the Materials,
 * either expressly, by implication, inducement, estoppel or
 * otherwise, except as expressly provided in the license
 * provided with the Materials
*/
// [DESCRIPTION]
// 
// Solve the nqueens problem with Win32 Threads
//
// [COMPILE]  
// Compile on IA-32 and Intel(R) 64 architectures only. 
// Use the following compiler options to compile: 
//
// CommandLine: icl /D=WIN32 [/D_USYNC-/D_DEBUG] nq-win32api.cpp winmm.lib
//  

#include <windows.h> // threading API
#include <iostream>  // std::cout, cin, cerr ...
#include <iomanip>   // modify std::out
#include <string>    // std::string
#include <sstream>   // std::stringstream

#include <mmsystem.h>

using namespace std;

#define _MULTI_THREADED
//#define _USYNC

// Thread stuff

const int n_threads=2;

#ifdef _USYNC
int nrOfUSolutions=0;
int nbrsol[n_threads] ;
#else
int nrOfSolutions=0;
#endif

int size=0;

CRITICAL_SECTION lock;  // declare a global CS object
CRITICAL_SECTION plock; // declare a global CS object

// Thread local data structure that will contain iteration range and message to print out
struct _thr_params {
  int start;
  int end;
  int id;
  int *queens;
  std::string msg;
};

typedef struct _thr_params thr_params;

void print(int queens[]) {
#ifndef _USYNC
  cerr << "Solution " << nrOfSolutions << endl; 
#endif
  for(int row=0; row<size; row++) {
    for(int col=0; col<size; col++) {
      if(queens[row]==col) {
        cout << "Q";
      }
      else {
        cout << "-";
      }
    }
    cout << endl;
  }
}

void setQueen(int queens[], int row, int col, int id) {
  int lid;

  lid=id;
  for(int i=0; i<row; i++) {
    // vertical attacks
    if (queens[i]==col) {
      return;
    }
    // diagonal attacks
    if (abs(queens[i]-col) == (row-i) ) {
      return;
    }
  }

  // column is ok, set the queen
  queens[row]=col;

  if(row==size-1) {

 #ifdef _USYNC
    nbrsol[lid]++ ;    
#else   
  //  InterlockedIncrement(&nrOfSolutions);
  EnterCriticalSection(&lock); // start critical section
  nrOfSolutions++;
  LeaveCriticalSection(&lock); // end of critical section
#endif

#ifdef _DEBUG
  EnterCriticalSection(&plock); // start critical section  
  // print(queens);
  LeaveCriticalSection(&plock); // end of critical section
#endif
  }
  else {
    // try to fill next row
    for(int i=0; i<size; i++) {
      setQueen(queens, row+1, i, lid);
    }
  }
}

// Worker thread function (to executed by each thread in parallel)
DWORD WINAPI run_solve (void* param) 
{
  // Retrieve arguments passed to this thread: param is a pointer to 
  // void, static_cast converts is to the pointer to thr_params
  thr_params* params = static_cast<thr_params*> (param); 

  // Print messages
  for (int i = params->start; i < params->end; ++i) {
    // create separate array for each recursion
#ifdef _DEBUG
    //cout << "Call set Queens with: row: 0 , column : " << i << endl;
    //std::stringstream s;
    //s << i; // convert i to stringstream
    // merge "static" part of the message with the "dynamic"
    //std::string message = params->msg + " solving row:" + s.str() + "\n";
    //std::cout << message;
#endif
    setQueen(new int[size], 0, i, params->id);

  }

  return 0;
}

void run_threaded_loop (int num_thr, int size, int _queens[]) {
  
  HANDLE* threads = new HANDLE[num_thr];
  thr_params* params = new thr_params[num_thr];

  for (int i = 0; i < num_thr; ++i) {
    // Give each thread equal number of rows
    params[i].start = i * (size/num_thr);
    params[i].end   = params[i].start + (size/num_thr);
    params[i].id     = i;  
    params[i].queens = _queens;
    
    // Construct a message:
    std::stringstream s;
    s << i; // make a string out of integer
    params[i].msg = "Thread #" + s.str();

    // Start a thread: it's important to pass argument-pointer to a different 
    // memory for each thread's parameter to avoid data races
    threads[i] = CreateThread (NULL, 0, run_solve, static_cast<void *> (&params[i]), 0, NULL);
  }

  // Join threads: wait until all threads are done
  WaitForMultipleObjects (num_thr, threads, true, INFINITE);

  // Free memory
  delete[] params;
  delete[] threads;
}

// -------------

int main(int argc, char*argv[]) {
  if(argc !=2) {
    cerr << "Usage: nq-winth boardSize.\n";
    return 0;
  }
  
  InitializeCriticalSection(&lock);  // allow a single thread access to the CS
  InitializeCriticalSection(&plock); // allow a single thread access to the CS
  
  size = atoi(argv[1]);
  
#ifdef _USYNC  
  for (int i = 0; i < n_threads; ++i) {
    nbrsol[i]=0;
  }
  cout << "Starting *unsynchronised* recursive win32 thread solver for size " << size << " with " << n_threads << " threads" << " ...\n";
#else
  cout << "Starting synchronised recursive win32 thread solver for size " << size << " with " << n_threads << " threads" << " ...\n";
#endif
  
  DWORD startTime=timeGetTime();
  run_threaded_loop (n_threads, size, new int[size]);
  DWORD endTime=timeGetTime();
  
#ifdef _USYNC
  for (int i = 0; i < n_threads; ++i) {
    nrOfUSolutions=nrOfUSolutions+nbrsol[i];
  }
  // Not synchronised solution
  cout << "Number of solutions: " << nrOfSolutions << endl;
  cout << "Calculations took: " << setw(10) << setprecision(2) << fixed << (endTime-startTime) << " ms." << endl;
#else
// Synchronised solution
  cout << "Number of solutions: " << nrOfSolutions << endl;
  cout << "Calculations took: " << (endTime-startTime) << " ms." << endl;
#endif  
  
  return 0;
}